/**
 * This code accompanies the paper:
 *
 * "Dense non-rigid shape correspondence using random forests"
 * E.Rodola, S.Rota Bulo, T.Windheuser, M.Vestner, D.Cremers.
 * Proc. CVPR 2014
 *
 * Please cite the paper above if you use the code.
 */
 
#include "ShapeRFClassifier.h"
#include <algorithm>

/**
 * Sort-by function. Wraps around std::sort().
 *
 * Example:
 *
 * struct A { ...; double measure; ... };
 *
 * std::vector<A> vec;
 * sort_by(vec.begin(), vec.end(), &A::measure);
 *
 * Written by Emanuele Rodolà.
 */

template <typename T, typename FieldPtr>
struct lt_field : public std::binary_function<T, T, bool>
{
    FieldPtr field_ptr;

    lt_field(FieldPtr field_ptr_) : field_ptr(field_ptr_) {}

    bool operator()(const T& a, const T& b) const
    {
        return (a.*field_ptr < b.*field_ptr);
    }
};

template <typename Iterator, typename FieldPtr>
void sort_by(Iterator begin, Iterator end, FieldPtr field_ptr)
{
   std::sort(
           begin,
           end,
           lt_field<typename std::iterator_traits<Iterator>::value_type, FieldPtr>(field_ptr));
}

/**
 * Converts a global index from the training set to a (shape,sample) pair.
 *
 * @param tidx      Global index from the whole set of training samples.
 * @param shapeIdx  Output index indicating the shape to which the given 'idx' belongs.
 * @param sampleIdx Output index indicating the shape _sample_ (which equals a point index in case all points are considered in the shape).
 */
void DataSet::getShapeSample(uint32_t tidx, uint32_t& shapeIdx, uint32_t& sampleIdx) const
{
   uint32_t accum = 0;

   for (uint32_t i=0; i<n_shapes; ++i)
   {
      const int sz = n_labels[i];

      if (tidx < accum+sz)
      {
         shapeIdx = i;
         sampleIdx = tidx-accum;
         break;
      }

      accum += sz;
   }

   if (accum == n_tot_labels)
      throw std::out_of_range("DataSet::getSample(): Given index is not valid for the current dataset.");
}

/**
 * Get the value of the k-th basis vector of a given shape computed at a given point (i.e. a scalar).
 *
 * @param shapeIdx    Input shape index in the training set.
 * @param sampleIdx   Input index indicating the shape _sample_ (which equals a point index in case all points are considered in the shape).
 * @param harmonicIdx Input index indicating the basis vector.
 */
double DataSet::getHarmonicValue(uint32_t shapeIdx, uint32_t sampleIdx, uint32_t harmonicIdx) const
{
   return shapes.at(shapeIdx)->getHarmonic(harmonicIdx,sampleIdx);
}

/**
 * Shape constructor. Precomputes data needed for computing descriptors.
 *
 * @param labels_    Labels on ALL points of the shape (not just at the desired samples).
 * @param samples_   Vector of point indices to consider in the training process.
 * @param evals_     Eigenvalues of the LB operator for the current shape.
 * @param evecs_     Eigenfunctions of the LB operator for the current shape. The eigenfunctions should
 *                   be defined at ALL points in the shape, not just at the desired samples.
 * @param max_nev    .
 * @param descr_size .
 */
Shape::Shape(
      const std::vector<int>& labels_,
      const std::vector<int>& samples_,
      const std::vector<double>& evals_,
      const std::vector<double>& evecs_,
      int max_nev,
      int descr_size
      )
   : evecs(evecs_), labels(labels_), samples(samples_), use_all_points(samples_.empty())
{
   // sort the eigen-pairs

   const int n_tot_eigenvalues = (int)evals_.size();
   const int n_total_points = (int)(evecs.size() / n_tot_eigenvalues);

   if (n_total_points < (int)samples.size())
      throw std::out_of_range("Shape(): the training process is given more samples than available points.");

   if (n_total_points != (int)labels.size())
      throw std::logic_error("Shape(): the training process must be given a label for each point in the training set.");

   eigenpairs.resize(n_tot_eigenvalues);
   for (int i=0; i<n_tot_eigenvalues; ++i)
   {
      eigenpairs[i].first = std::fabs(evals_[i]);
      eigenpairs[i].second = &evecs[ i*n_total_points ];
   }

   sort_by(eigenpairs.begin(), eigenpairs.end(), &std::pair<double,const double*>::first);

   //std::cout << "Precomputing Gaussians... " << std::flush;

   // the energy distribution of E_k is modeled as a normal distribution with
   // mean = log(E_k) and variance = sigma^2 (wave kernel signature)

   std::vector<double> logE(n_tot_eigenvalues);

   for (int i=0; i<n_tot_eigenvalues; ++i)
      logE[i] = std::log( std::max(1e-6, eigenpairs[i].first) );

   //TODO: this could be made a parameter as well
   const double variance = 6.0;

   const int descriptor_size = 1 + descr_size;
   const int first_n_eigenvalues = 1 + max_nev;

   wks_gaussians.resize(descriptor_size);
   for (int i=0; i<descriptor_size; ++i)
   {
      wks_gaussians[i].resize(first_n_eigenvalues);
      for (int nev=0; nev<first_n_eigenvalues; ++nev)
         wks_gaussians[i][nev].resize(nev);
   }

   const double e_min = logE[1];

   // energy measurements will be performed at each point for 'n_evaluations' equally-spaced
   // energy scales in [e_min,e_max].

   for (int energy_level = 0; energy_level < descriptor_size; ++energy_level)
   {
      for (int nev = 0; nev < first_n_eigenvalues; ++nev)
      {
         const double e_max = logE[nev-1]/1.02;
         const double e_step = (e_max-e_min)/(descriptor_size-1);
         const double energy = e_min + energy_level*e_step;
         const double sigma = e_step*variance;
         const double den = 2.*sigma*sigma;

         for (int k=0; k<nev; ++k)
         {
            const double num = energy-logE[k];
            const double gaussian = std::exp( - (num*num) / den );
            wks_gaussians[energy_level][nev][k] = gaussian;
         }
      } // next eigenfunction
   } // next energy level

   //std::cout << "done." << std::endl;
}

/**
 * Computes a single channel of the wave kernel signature at a given point.
 *
 * @param point_idx     Point index in the shape.
 * @param n_eigenvalues Total number of eigenvalue/eigenvector pairs to use.
 * @param energy_level  Energy level of the measurement (one component of the wave kernel signature).
 */
double Shape::compute_wks(int point_idx, int n_eigenvalues, int energy_level) const
{
   double C = 0.; // normalization weight
   double wks = 0.;

   const std::vector<double>& G = wks_gaussians.at(energy_level).at(n_eigenvalues);

   for (int k=0; k<n_eigenvalues; ++k)
   {
      const double gaussian = G[k];
      const double phi = eigenpairs[k].second[point_idx];

      wks += phi*phi*gaussian;
      C += gaussian;
   }

   return wks / C;
}
